"""Metadata inspector."""
from __future__ import annotations
import argparse
from functools import partial
from pathlib import Path
import warnings
import sys
from typing import Optional, TextIO

from cftime import num2date
from dask import array as dask_array
from hurry.filesize import alternative, size
import numpy as np
import xarray as xr
import yaml

from ._version import __version__
from ._slk import get_slk_metadata, login


def _summarize_datavar(name: str, var: xr.DataArray, col_width: int) -> str:
    out = [
        xr.core.formatting.summarize_variable(name, var.variable, col_width)
    ]
    if var.attrs:
        n_spaces = 0
        for k in out[0]:
            if k == " ":
                n_spaces += 1
            else:
                break
        out += [
            n_spaces * " " + line
            for line in xr.core.formatting.attrs_repr(var.attrs).split("\n")
        ]
    return "\n".join(out)


def parse_args(args: Optional[list[str]] = None) -> tuple[list[Path], bool]:
    """Construct command line argument parser."""

    argp = argparse.ArgumentParser
    app = argp(
        prog="metadata-inspector",
        description=(
            "Inspect meta data of a weather/climate datasets "
            "with help of xarray"
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    app.add_argument(
        "input",
        metavar="input",
        type=Path,
        nargs="+",
        help="Input files that will be processed",
    )
    app.add_argument(
        "--html",
        action="store_true",
        help="Create html representation of the dataset.",
    )
    app.add_argument(
        "--version",
        "-V",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    parsed_args = app.parse_args(args)
    return parsed_args.input, parsed_args.html


def dataset_from_hsm(input_file: str) -> xr.Dataset:
    """Create a dataset view from attributes."""

    attrs: dict[str, dict[str, str]] = (
        yaml.safe_load(get_slk_metadata(input_file)) or {}
    ).get("Keywords", {})

    dset = xr.Dataset({}, attrs=attrs.pop("global"))
    for dim in attrs.pop("dims"):
        size = int(attrs[dim].pop("size"))
        start, end = float(attrs[dim].pop("start")), float(
            attrs[dim].pop("end")
        )
        vec = np.linspace(start, end, size)
        if dim == "time":
            vec = num2date(vec, attrs[dim]["units"], attrs[dim]["calendar"])
        dset[dim] = xr.DataArray(vec, name=dim, dims=(dim,), attrs=attrs[dim])
    for data_var in attrs.pop("data_vars"):
        dims = attrs[data_var].pop("dims")
        sizes = [dset[d].size for d in dims]
        dset[data_var] = xr.DataArray(
            dask_array.empty(sizes),
            name=data_var,
            dims=dims,
            attrs=attrs[data_var],
        )
    return dset


def _get_files(input_: list[Path]) -> tuple[list[str], list[str]]:
    """Get all files from given input"""

    files_fs: list[str] = []
    files_archive: list[str] = []
    extensions: tuple[str, ...] = (
        ".nc",
        ".nc4",
        ".grb",
        ".grib",
        ".grib2",
        ".grb2",
        ".zarr",
        ".h5",
        ".hdf5",
    )
    for inp_file in input_:
        inp = inp_file.expanduser().absolute()
        if inp.is_dir() and inp.exists():
            files_fs += [
                str(inp_file)
                for inp_file in inp.rglob("*")
                if inp_file.suffix.lower() in extensions
            ]
        elif inp.is_file() and inp.exists():
            files_fs.append(str(inp))
        elif inp.parent.exists() and inp.parent.is_dir():
            files_fs += [
                str(inp_file)
                for inp_file in inp.parent.rglob(inp.name)
                if inp_file.suffix in extensions
            ]
        elif inp.parts[1] == "arch" or str(inp).startswith("slk:"):
            files_archive.append(str(inp))
    return sorted(files_fs), sorted(files_archive)


def _open_datasets(files_fs: list[str], files_hsm: list[str]) -> xr.Dataset:
    dsets: list[xr.Dataset] = []
    if files_fs:
        dsets.append(
            xr.open_mfdataset(
                files_fs, parallel=False, combine="by_coords", use_cftime=True
            )
        )
    if files_hsm:
        login()
        for inp_file in files_hsm:
            dsets.append(dataset_from_hsm(inp_file))
    return xr.merge(dsets)


def main(input_files: list[Path], html: bool = False) -> tuple[str, TextIO]:
    """Print the representation of a dataset.

    Parameters
    ----------

    input_files: list[Path]
        Collection of input files that are to be inspected
    html: bool, default: True
        If true a representation suitable for html is displayed.
    """

    files_fs, files_hsm = _get_files(input_files)
    if not files_fs and not files_hsm:
        return "No files found", sys.stderr
    try:
        dset = _open_datasets(files_fs, files_hsm)
    except Exception as error:
        error_header = (
            "No data found, file(s) might be corrupted. "
            "See err. message below:"
        )
        error_msg = str(error)
        if html:
            error_msg = error_msg.replace("\n", "<br>")
            msg = (
                "<p><b>Error:</b>Could not open dataset for more details "
                "do not use the --html flag.</p>"
            )
            return msg, sys.stdout

        else:
            msg = f"{error_header}\n{error_msg}"
            return msg, sys.stderr
    fsize = size(dset.nbytes, system=alternative)
    if html:
        out_str = xr.core.formatting_html.dataset_repr(dset)
    else:
        xr.core.options.OPTIONS["display_expand_data_vars"] = True
        xr.core.options.OPTIONS["display_expand_attrs"] = True
        xr.core.options.OPTIONS["display_expand_data"] = True
        xr.core.options.OPTIONS["display_max_rows"] = 100
        xr.core.formatting.data_vars_repr = partial(
            xr.core.formatting._mapping_repr,
            title="Data variables",
            summarizer=_summarize_datavar,
            expand_option_name="display_expand_data_vars",
        )
        out_str = xr.core.formatting.dataset_repr(dset)
    replace_str = (
        ("xarray.Dataset", f"Dataset (byte-size: {fsize})"),
        (
            "<svg class='icon xr-icon-file-text2'>",
            "<i class='fa fa-file-text-o'>",
        ),
        ("<svg class='icon xr-icon-database'>", "<i class='fa fa-database'>"),
        ("</use></svg>", "</use></i>"),
        ("numpy.", ""),
        ("np.", ""),
        ("dask.", ""),
    )
    for entry, replace in replace_str:
        out_str = out_str.replace(entry, replace)

    return out_str, sys.stdout


def cli(args: Optional[list[str]] = None) -> None:
    """Command line argument inteface."""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        try:
            msg, text_io = main(*parse_args(args))
        except Exception as error:
            msg, text_io = f"Error: {error}", sys.stderr
    print(msg, file=text_io, flush=True)



"""Module that interacts with slk.

Until we do have access to the rest api we'll have to rely on the command line
tools.
"""


from __future__ import annotations
import base64
from datetime import datetime, timedelta
from getpass import getuser
import json
from pathlib import Path
import os
import warnings
from subprocess import run, PIPE, SubprocessError

import requests

SLK_PATH = "/sw/spack-levante/slk-3.3.21-5xnsgp/bin"
JDK_PATH = "/sw/spack-levante/openjdk-17.0.0_35-k5o6dr/bin"
JAVA_HOME = "/sw/spack-levante/openjdk-17.0.0_35-k5o6dr"

SESSION_PATH = Path("~").expanduser() / ".slk" / "config.json"


def get_env() -> dict[str, str]:
    """Prepare the environment variables for slk."""

    env = os.environ.copy()
    env["PATH"] = f"{SLK_PATH}:{env['PATH']}"
    env["PATH"] = f"{JDK_PATH}:{env['PATH']}"
    env["JAVA_HOME"] = "{JAVA_HOME}"
    return env


def get_slk_metadata(input_path: str) -> str:
    """Extract dataset metdata from path in the hsm.

    Parameters
    ----------
    input_path: Path
        The hsm path the metdata is extracted from


    Returns
    -------
    str: string representation of the metdata
    """
    command = ["slk_helpers", "metadata", input_path]
    try:
        res = run(command, env=get_env(), check=True, stdout=PIPE, stderr=PIPE)
    except SubprocessError as error:  # pragma: no cover
        warnings.warn(
            f"Error: could not get metdata: {error}"
        )  # pragma: no cover
        return ""  # pragma: no cover
    lines: list[str] = []
    # This needs to be done because the output of the command is only nearly
    # yaml. That is the ":" for the first keys are missing:
    # For example:
    # document
    #      Version: foo
    # netcdf
    #     id: bar
    #     var_name: tas
    # Since yaml needs could not handle this we have to add the ':' to the
    # keys manually.
    for line in [o.strip() for o in res.stdout.decode().split("\n")]:
        if line.strip().lower().startswith("keywords:"):
            lines.append(line)
    return "\n".join(lines)


def get_expiration_date() -> datetime:
    """Get the expiration date of the session key."""
    session_path = Path("~").expanduser() / ".slk" / "config.json"
    fmt = "%a %b %d %H:%M:%S %Z %Y"
    now = datetime.now().astimezone().strftime(fmt)
    try:
        with session_path.open() as f_obj:
            date = json.load(f_obj).get("expireDate", now)
    except FileNotFoundError:  # pragma: no cover
        date = now  # pragma: no cover
    return datetime.strptime(date, fmt)


def _login_via_request(passwd: str) -> None:
    data = {
        "data": {
            "attributes": {
                "domain": "ldap",
                "name": getuser(),
                "password": passwd,
            },
            "type": "authentication",
        }
    }
    headers = {"Content-type": "application/json"}
    fmt = "%a %b %d %H:%M:%S %Z %Y"
    exp_date = (datetime.now() + timedelta(days=20)).astimezone().strftime(fmt)
    url = "https://archive.dkrz.de/api/v2/authentication"
    res = requests.post(
        url, data=json.dumps(data), headers=headers, verify=False
    )
    key = (
        res.json().get("data", {}).get("attributes", {}).get("session_key", "")
    )
    if key:
        sec = {"user": getuser(), "sessionKey": key, "expireDate": exp_date}
        SESSION_PATH.parent.mkdir(exist_ok=True, parents=True)
        with SESSION_PATH.open("w") as f_obj:
            json.dump(sec, f_obj)
        SESSION_PATH.chmod(0o600)


def login() -> None:
    """Login to the system."""
    passwd = os.environ.get("LC_TELEPHONE", "")
    now = datetime.now()
    exp_date = get_expiration_date()
    diff = (exp_date - now).total_seconds()
    if passwd:
        passwd = base64.b64decode(passwd.encode()).decode()
        _login_via_request(passwd)
    elif diff <= 0:
        print("Your session has expired, login to slk")
        run(["slk", "login"], shell=False, check=True, env=get_env())



#!/usr/bin/env python3
"""Setup script for packaging metadata-inspektor."""

from pathlib import Path
import re
from setuptools import setup, find_packages


def read(*parts):
    """Read the content of a file."""
    script_path = Path(__file__).parent
    with script_path.joinpath(*parts).open() as f:
        return f.read()


def find_version(*parts):
    """The the version in a given file."""
    vers_file = read(*parts)
    match = re.search(r'^__version__ = "(\d+.\d+.\d+)"', vers_file, re.M)
    if match is not None:
        return match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="metadata_inspector",
    version=find_version("src", "metadata_inspector", "_version.py"),
    author="Martin Bergemann",
    author_email="bergemann@dkrz.de",
    maintainer="Martin Bergemann",
    url="https://github.com/FREVA-CLINT/metadata-inspektor.git",
    description="Inspect metadata of weather/climate datasets",
    long_description=read("README.md"),
    license="BSD-3-Clause",
    long_description_content_type="text/markdown",
    include_package_data=True,
    project_urls={
        "Source": "https://github.com/FREVA-CLINT/metadata-inspektor",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
    ],
    packages=find_packages("src"),
    package_dir={"": "src"},
    entry_points={"console_scripts": ["metadata-inspector = metadata_inspector:cli"]},
    install_requires=[
        "cftime",
        "dask[diagnostics]",
        "hurry.filesize",
        "h5netcdf",
        "netCDF4",
        "requests",
        "xarray",
    ],
    extras_require={
        "tests": [
            "black",
            "pytest",
            "pandas",
            "mock",
            "numpy",
            "requests-mock",
            "pytest-env",
            "pytest-cov",
            "testpath",
            "flake8",
            "mypy",
            "types-mock",
            "types-PyYAML",
            "types-requests",
            "types-setuptools",
        ],
    },
    python_requires=">=3.7",
)
